/**************************************************************************
*  Copyright (c) 2015 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  22.09.2015  mifi  First Version, tested with an Olimex STM32-P405 board.
**************************************************************************/
#define __MAIN_C__

/*=======================================================================*/
/*  Includes                                                             */
/*=======================================================================*/
#include <stdio.h>
#include "tal.h"
#include "terminal.h"
#include "sht15.h"
#include "bmp085.h"

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

union CONVERT
{
   float   fValue;
   uint8_t bArray[4];
};

#define MY_NODE_ID   1

/*=======================================================================*/
/*  Definition of all global Data                                        */
/*=======================================================================*/

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

/* 
 * Some TASK variables like stack and task control block.
 */
static OS_STACK (StartStack,   TASK_START_STK_SIZE);
static OS_STACK (MeasureStack, TASK_MEASURE_STK_SIZE);

static os_tcb_t TCBStartTask;
static os_tcb_t TCBMeasureTask;

static TAL_CAN_OBJECT   RxRingBuffer[8];
static TAL_CAN_OBJECT   TxRingBuffer[8];

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*************************************************************************/
/*  OutputBootMessage                                                    */
/*                                                                       */
/*  Output boot message.                                                 */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void OutputBootMessage (void)
{
   const char ResetScreen[] = { 0x1B, 'c', 0 };
   
   term_printf("%s", ResetScreen);
   OSTimeDly(10);   
   
   term_printf("\n");
   term_printf("*********************************\n");
   term_printf("  Project: %s\n", PROJECT_NAME);
   term_printf("  Board  : %s\n", TAL_BOARD);
   term_printf("  Version: v%s\n", PROJECT_VER_STRING);
   term_printf("  Build  : "__DATE__ " " __TIME__"\n");
   term_printf("*********************************\n");
   term_printf("\n");
} /* OutputBootMessage */

/*************************************************************************/
/*  OutputVersionInfo                                                    */
/*                                                                       */
/*  Output version information.                                          */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void OutputVersionInfo (void)
{
   term_printf("*** Version ***\n");
   term_printf("OS   : v%s\n", OS_VER_STRING);
   term_printf("TAL  : v%s\n", TAL_CORE_VER_STRING);
   term_printf("Board: %s\n", TAL_BOARD);
   term_printf("CPU  : %s\n", TAL_CPU);
   term_printf("\n");
} /* OutputVersionInfo */

/*************************************************************************/
/*  OutputFrequencyInfo                                                  */
/*                                                                       */
/*  Output frequency information.                                        */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void OutputFrequencyInfo (void)
{
   term_printf("*** Frequency info ***\n");
   term_printf("CPU : %3d MHz\n", tal_CPUGetFrequencyCPU() / 1000000); 
   term_printf("AHB : %3d MHz\n", tal_CPUGetFrequencyAHB2() / 1000000); 
   term_printf("APB1: %3d MHz\n", tal_CPUGetFrequencyAPB1() / 1000000); 
   term_printf("APB2: %3d MHz\n", tal_CPUGetFrequencyAPB2() / 1000000); 
   term_printf("HSE : %3d MHz\n", HSE_VALUE / 1000000); 
   term_printf("\n");
} /* OutputFrequencyInfo */

/*************************************************************************/
/*  OutputCPULoad                                                        */
/*                                                                       */
/*  Output load information.                                             */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void OutputCPULoad (void)
{
   term_printf("CPU load: %d%%\n\n", OSStatGetCPULoad());
} /* OutputCPULoad */

/*************************************************************************/
/*  OutputUsageInfo                                                      */
/*                                                                       */
/*  Output usage information.                                            */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void OutputUsageInfo (void)
{
   term_printf("*** Usage ***\n");
   term_printf("h: output this info\n");
   term_printf("?: output this info\n");
   term_printf("c: output cpu load info\n");
   term_printf("f: output frequency info\n");
   term_printf("r: output runtime stack info\n");
   term_printf("t: output task info\n");
   term_printf("v: output version info\n");
   term_printf("\n");
} /* OutputUsageInfo */

/*************************************************************************/
/*  MeasureTask                                                          */
/*                                                                       */
/*  In    : task parameter                                               */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/
static void MeasureTask (void *p)
{
   float            fTemperature1;
   float            fTemperature2;
   uint8_t          bHumidity;
   uint16_t         wPressure;
   uint32_t         dStartTime;
   uint32_t         dEndTime;
   union CONVERT     Convert;
   TAL_CAN_DCB       CANPort;
   TAL_CAN_SETTINGS  CANSettings;   
   TAL_CAN_OBJECT    CANObject;

   (void)p;
   
   /* Init the Device Control Block */      
   tal_CANInitDCB(&CANPort, TAL_CAN_PORT_1);
   
   /* Set Rx and Tx ring buffer */
   tal_CANSetRingBuffer(&CANPort, TAL_CAN_BUFFER_RX, RxRingBuffer, sizeof(RxRingBuffer)); 
   tal_CANSetRingBuffer(&CANPort, TAL_CAN_BUFFER_TX, TxRingBuffer, sizeof(TxRingBuffer)); 

   /* Prepare communication settings */
   CANSettings.dBitRate = TAL_CAN_BR_500K; 
   
   /* Now, the port can be opened */   
   tal_CANOpen(&CANPort, &CANSettings);   

   while(1)
   {
      dStartTime = OSTimeGet();
      
      sht15_Measurement(&fTemperature1, &bHumidity);
      bmp085_Measurement(0, &fTemperature2, &wPressure);
      
      term_printf("SHT15  Temp    : %.1f\n", fTemperature1); 
      term_printf("BMP085 Temp    : %.1f\n", fTemperature2); 
      term_printf("SHT15  Humidity: %d\n", bHumidity); 
      term_printf("BMP085 Pressure: %d\n", wPressure); 
      term_printf("\n", wPressure); 

      /* 
       * Transmit SHT15 data 
       */
      Convert.fValue = fTemperature1;
      
      CANObject.dIdentifier = 0x180 + MY_NODE_ID;
      CANObject.bFlags      = TAL_CAN_OBJECT_FLAGS_STD_ID;
      CANObject.bDLC        = 8;
      CANObject.Data[0]     = Convert.bArray[0];
      CANObject.Data[1]     = Convert.bArray[1];
      CANObject.Data[2]     = Convert.bArray[2];
      CANObject.Data[3]     = Convert.bArray[3];
      CANObject.Data[4]     = bHumidity;
      CANObject.Data[5]     = 0;
      CANObject.Data[6]     = 0;
      CANObject.Data[7]     = 0;
      
      tal_CANMsgSend(&CANPort, &CANObject);


      /* 
       * Transmit BMP085 data 
       */
      Convert.fValue = fTemperature2;

      CANObject.dIdentifier = 0x280 + MY_NODE_ID;
      CANObject.bFlags      = TAL_CAN_OBJECT_FLAGS_STD_ID;
      CANObject.bDLC        = 8;
      CANObject.Data[0]     = Convert.bArray[0];
      CANObject.Data[1]     = Convert.bArray[1];
      CANObject.Data[2]     = Convert.bArray[2];
      CANObject.Data[3]     = Convert.bArray[3];
      CANObject.Data[4]     = (uint8_t)(wPressure & 0x00FF);;
      CANObject.Data[5]     = (uint8_t)((wPressure >> 8) & 0x00FF);
      CANObject.Data[6]     = 0;
      CANObject.Data[7]     = 0;
      
      tal_CANMsgSend(&CANPort, &CANObject);
      
      dEndTime = OSTimeGet();
      

      /*
       * OSTimeDlyUntil is not available in the moment,
       * therefore OSTimeDly must be used.
       */
      OSTimeDly(TASK_MEASURE_DELAY_MS - (dEndTime - dStartTime));
//      OSTimeDlyUntil(dStartTime + TASK_MEASURE_DELAY_MS);
   }
} /* MeasureTask */

/*************************************************************************/
/*  StartTask                                                            */
/*                                                                       */
/*  This is the Start task.                                              */
/*                                                                       */
/*  In    : task parameter                                               */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/
static void StartTask (void *p)
{
   (void)p;
   
   /*
    * The StartTask will be used to start all other tasks in the system.
    * At the end the priority will be set to "IDLE" priority.
    */

   /* 
    * Init the statistic function 
    */
   OSStatEnable();
   
   /*
    * Star the Terminal functionality 
    */
   term_Start();
   
   /*
    * Output startup messages
    */
   OutputBootMessage();
   OutputUsageInfo();
   
   sht15_Init();
   bmp085_Init();
   
   /* Create the Measure task */
   OSTaskCreate(&TCBMeasureTask, MeasureTask, NULL, TASK_MEASURE_PRIORITY,
                MeasureStack, sizeof(MeasureStack), 
                "MeasureTask");

   OSTaskChangePriority(TASK_START_PRIORITY_IDLE);

   while (1)
   {
      tal_LEDSet(TAL_LED_CHANNEL_1);
      OSTimeDly(TASK_START_DELAY_MS);
      tal_LEDClear(TAL_LED_CHANNEL_1);
      OSTimeDly(TASK_START_DELAY_MS);
   }
} /* StartTask */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  main                                                                 */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: never                                                        */
/*************************************************************************/
int main (void)
{
   /* 
    * Init the Tiny Abstraction Layer 
    */
   tal_Init();
   
   /* Create the Start task */
   OSTaskCreate(&TCBStartTask, StartTask, NULL, TASK_START_PRIORITY,
                StartStack, sizeof(StartStack), 
                "StartTask");

   /* 
    * OSStart must be the last function here.
    *  
    * Fasten your seatbelt, start the engine...
    */        
   OSStart();
   
   /*
    * This return here make no sense.
    * But to prevent the compiler warning:
    *    "return type of 'main' is not 'int'
    * We use an int as return :-)
    */ 
   return(0); /*lint !e527*/
} /* main */

/*************************************************************************/
/*  term_RxCallback                                                      */
/*                                                                       */
/*  Will be called from TermTask in case a char is received.             */
/*                                                                       */
/*  In    : bData                                                        */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void term_RxCallback (uint8_t bData)
{
   switch (bData)
   {
      case 'h':
      case '?':
      {
         OutputUsageInfo();
         break;
      } 
           
      case 'c':
      {
         OutputCPULoad();
         break;
      }

      case 'f':
      {
         OutputFrequencyInfo();
         break;
      }

      case 'r':
      {
         OSOutputRuntimeStackInfo();
         break;
      }
      
      case 't':
      {
         OSOutputTaskInfo();
         break;
      }
      
      case 'v':
      {
         OutputVersionInfo();
         break;
      }
      
      default:
      {
         /* Do nothing */
         break;
      }
   } /* end switch (bData) */
} /* term_RxCallback */

/*** EOF ***/
