/**************************************************************************
*  Copyright (c) 2011-2015 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  06.08.2011  mifi  First Version
*  19.04.2013  mifi  Remove unused code.
**************************************************************************/
#define __SHT15_C__

/*
 * All the information are based on the following documents:
 * 
 * Datasheet SHT1x (SHT10, SHT11, SHT15) Humidity and Temperature Sensor
 * Version 4.3 - May 2011
 *
 * SHT1x and SHT7x Sample Code Humidity & Temperature Sensor
 * Version 2.07 - November 2011
 *
 * Application Note SHTxx Humidity & Temperature Sensor
 * Version 1.1 - October 2008
 *
 * For more information take a look at: www.sensirion.com
 */
  
/*=======================================================================*/
/*  Include                                                              */
/*=======================================================================*/
#include <stdint.h>
#include "tal.h"
#include "sht15.h"

/*=======================================================================*/
/*  All Structures and Common Constants                                  */
/*=======================================================================*/

/* Use PB0 for clock */
#define CLK_LOW()    { GPIOB->BSRRH = (1<<0); }
#define CLK_HIGH()   { GPIOB->BSRRL = (1<<0); }

/* Use PB1 for data */
#define DATA_LOW()   { GPIOB->BSRRH = (1<<1); }
#define DATA_HIGH()  { GPIOB->BSRRL = (1<<1); }
#define DATA_GET()   ((GPIOB->IDR & (1<<1)) ? 1 : 0)

#define WAIT_HIGH()  { OSTimeDly(2); }
#define WAIT_LOW()   { OSTimeDly(2); }
#define WAIT()       { OSTimeDly(2); }

#define CNT_TIMEOUT  5000000L

/*
 * Some defines for the communication with the sht15
 */
#define ACK                1
#define NO_ACK             0

#define CMD_TEMP           0x03
#define CMD_HUMI           0x05
#define CMD_STATUS_REG_W   0x06
#define CMD_STATUS_REG_R   0x07


/*=======================================================================*/
/*  Prototypes                                                           */
/*=======================================================================*/

/*=======================================================================*/
/*  Definition of all local Data                                         */
/*=======================================================================*/

/*
 * 256 byte CRC Lookup table
 */
static const uint8_t CRCTable[256] = 
{
   0, 49, 98, 83, 196, 245, 166, 151, 185, 136, 219, 234, 125, 76, 31, 46, 
   67, 114, 33, 16, 135, 182, 229, 212, 250, 203, 152, 169, 62, 15, 92, 109, 
   134, 183, 228, 213, 66, 115, 32, 17, 63, 14, 93, 108, 251, 202, 153, 168,
   197, 244, 167, 150, 1, 48, 99, 82, 124, 77, 30, 47, 184, 137, 218, 235, 61, 
   12, 95, 110, 249, 200, 155, 170, 132, 181, 230, 215, 64, 113, 34, 19, 126, 
   79, 28, 45, 186, 139, 216, 233, 199, 246, 165, 148, 3, 50, 97, 80, 187,
   138, 217, 232, 127, 78, 29, 44, 2, 51, 96, 81, 198, 247, 164, 149, 248,
   201, 154, 171, 60, 13, 94, 111, 65, 112, 35, 18, 133, 180, 231, 214, 122,
   75, 24, 41, 190, 143, 220, 237, 195, 242, 161, 144, 7, 54, 101, 84, 57,
   8, 91, 106, 253, 204, 159, 174, 128, 177, 226, 211, 68, 117, 38, 23, 252,
   205, 158, 175, 56, 9, 90, 107, 69, 116, 39, 22, 129, 176, 227, 210, 191,
   142, 221, 236, 123, 74, 25, 40, 6, 55, 100, 85, 194, 243, 160, 145, 71,
   118, 37, 20, 131, 178, 225, 208, 254, 207, 156, 173, 58, 11, 88, 105, 4,
   53, 102, 87, 192, 241, 162, 147, 189, 140, 223, 238, 121, 72, 27, 42,
   193, 240, 163, 146, 5, 52, 103, 86, 120, 73, 26, 43, 188, 141, 222, 239,
   130, 179, 224, 209, 70, 119, 36, 21, 59, 10, 89, 104, 255, 206, 157, 172
};   

/*=======================================================================*/
/*  Definition of all local Procedures                                   */
/*=======================================================================*/

/*************************************************************************/
/*  ReverseData                                                          */
/*                                                                       */
/*  Bit 7 from the input data is bit 0 from the output data.             */
/*  Bit 6 from the input data is bit 1 from the output data.             */
/*  And so on ...                                                        */
/*                                                                       */
/*  In    : Data                                                         */
/*  Out   : none                                                         */
/*  Return: Bit reversed data                                            */
/*************************************************************************/
static uint8_t ReverseData (uint8_t Data)
{
   uint8_t RevData = 0;
   
   if (Data & 0x80) RevData |= 0x01;
   if (Data & 0x40) RevData |= 0x02;
   if (Data & 0x20) RevData |= 0x04;
   if (Data & 0x10) RevData |= 0x08;
   if (Data & 0x08) RevData |= 0x10;
   if (Data & 0x04) RevData |= 0x20;
   if (Data & 0x02) RevData |= 0x40;
   if (Data & 0x01) RevData |= 0x80;
   
   return(RevData);
} /* ReverseData */

/*************************************************************************/
/*  HWInit                                                               */
/*                                                                       */
/*  Setup the port pins like clk and data.                               */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void HWInit (void)
{
   /* Turn on GPIOB */
   RCC->AHB1ENR |= RCC_AHB1ENR_GPIOBEN;

   /* Set PB0 to output (CLK) */
   GPIOB->MODER &= ~GPIO_MODER_MODER0;
   GPIOB->MODER |= GPIO_MODER_MODER0_0;
   
   CLK_LOW();
   
   /* Set PB1 to output, Open-drain (DATA) */
   GPIOB->MODER  &= ~GPIO_MODER_MODER1;
   GPIOB->MODER  |= GPIO_MODER_MODER1_0;
   GPIOB->OTYPER |= GPIO_OTYPER_OT_1;
   
   DATA_LOW();   
} /* HWInit */

/*************************************************************************/
/*  TransStart                                                           */
/*                                                                       */
/*  Sending a "Transmission Start" sequence.                             */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void TransStart (void)
{
   DATA_HIGH();
   CLK_LOW();
   WAIT();
   
   CLK_HIGH();
   WAIT();
   DATA_LOW();
   WAIT();
   CLK_LOW();
   WAIT();
   WAIT();
   CLK_HIGH();
   WAIT();
   DATA_HIGH();
   WAIT();
   CLK_LOW();
   WAIT();
   WAIT();
} /* TransStart */

/*************************************************************************/
/*  COMReset                                                             */
/*                                                                       */
/*  Connection reset sequence. It will generate high / low clocks with   */
/*  about 570ns running from flash at 72MHz.                             */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
static void COMReset (void)
{
   uint8_t i;
   
   DATA_HIGH();
   CLK_LOW();
   
   for(i=0; i<9; i++)
   {
      CLK_HIGH();
      WAIT_HIGH();
      
      CLK_LOW();
      WAIT_LOW();
   }
   
   TransStart();
} /* COMReset */

/*************************************************************************/
/*  WaitFinished                                                         */
/*                                                                       */
/*  Wait until sensor has finished the measurement.                      */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: 0 = no error / 1 = error                                     */
/*************************************************************************/
static uint8_t WaitFinished (void)
{
   uint32_t Counter;
   uint8_t  Error = 0;
   
   Counter = 0;
   while(Counter < CNT_TIMEOUT)
   {
      if(DATA_GET() == 0)
      {
         break;
      }
      Counter++;
   }
   
   /* Check timeout */
   if (Counter >= CNT_TIMEOUT)
   {
      Error = 1;
   }

   return(Error);
} /* WaitFinished */

/*************************************************************************/
/*  WriteByte                                                            */
/*                                                                       */
/*  Writes a byte on the Sensibus and checks the acknowledge.            */
/*                                                                       */
/*  In    : Value                                                        */
/*  Out   : none                                                         */
/*  Return: 0 = no error / 1 = error                                     */
/*************************************************************************/
static uint8_t WriteByte (uint8_t Value)
{
   uint8_t i;
   uint8_t Error;
   
   for(i=0x80; i>0; i=i >> 1)
   {
      if (Value & i)
      {
         DATA_HIGH();
      }
      else
      {
         DATA_LOW();
      }
      WAIT();
      
      CLK_HIGH();
      WAIT();
      WAIT();
      CLK_LOW();
      WAIT();
   }
   
   /* Release DATA-line */
   DATA_HIGH();
   WAIT();
   
   CLK_HIGH();
   WAIT();
   
   /* Check ack (DATA will be pulled down by SHT1x) */
   Error = DATA_GET();
   
   CLK_LOW();
   WAIT();

   return(Error);
} /* WriteByte */

/*************************************************************************/
/*  ReadByte                                                             */
/*                                                                       */
/*  Reads a byte form the Sensibus and gives an acknowledge in case      */
/*  of CheckACK == 1                                                     */
/*                                                                       */
/*  In    : CheckACK                                                     */
/*  Out   : none                                                         */
/*  Return: Value                                                        */
/*************************************************************************/
static uint8_t ReadByte (uint8_t CheckACK)
{
   uint8_t i;
   uint8_t Value = 0;
   
   /* Release DATA-line */
   DATA_HIGH();
   
   /* Read value */
   for(i=0x80; i>0; i = i >> 1)
   {
      CLK_HIGH();
      WAIT();
      if (DATA_GET())
      {
         Value = Value | i;
      }
      CLK_LOW();
      WAIT();
   }
   
   /* Check ACK must set */
   if (1 == CheckACK)
   {
      DATA_LOW();
   }
   WAIT();
   CLK_HIGH();
   WAIT();
   WAIT();
   CLK_LOW();
   WAIT();
   
   /* Release DATA-line */
   DATA_HIGH();
   WAIT();
   
   return(Value);
} /* ReadByte */

/*************************************************************************/
/*  ReadData                                                             */
/*                                                                       */
/*  Read the temperature or humidity.                                    */
/*                                                                       */
/*  In    : CMD, pData                                                   */
/*  Out   : pData                                                        */
/*  Return: 0 = no error / 1 = error                                     */
/*************************************************************************/
static uint8_t ReadData (uint8_t CMD, uint16_t *pData)
{
   uint8_t  Error = 0;
   uint8_t  LowByte;
   uint8_t  HighByte;
   uint8_t  Checksum;
   uint8_t  CalcCRC = 0;
   
   TransStart();
   Error |= WriteByte(CMD);
   Error |= WaitFinished();

   HighByte = ReadByte(ACK);
   LowByte  = ReadByte(ACK);
   Checksum = ReadByte(NO_ACK);
   Checksum = ReverseData(Checksum);
   
   /* Test checksum */
   CalcCRC = CRCTable[CalcCRC ^ CMD];
   CalcCRC = CRCTable[CalcCRC ^ HighByte];
   CalcCRC = CRCTable[CalcCRC ^ LowByte];
   if (CalcCRC != Checksum)
   {
      Error = 1;
   } 
   
   *pData = ((uint16_t)HighByte << 8) | LowByte;
   
   return(Error);
} /* ReadData */

#if 0
/*************************************************************************/
/*  ReadStatus                                                           */
/*                                                                       */
/*  Reads the status register.                                           */
/*                                                                       */
/*  In    : pValue                                                       */
/*  Out   : pValue                                                       */
/*  Return: 0 = no error / 1 = error                                     */
/*************************************************************************/
static uint8_t ReadStatus (uint8_t *pValue)
{
   uint8_t Error;
   uint8_t Value;
   uint8_t Checksum;
   uint8_t CalcCRC = 0;
   
   TransStart();
   Error    = WriteByte(CMD_STATUS_REG_R);
   Value    = ReadByte(ACK);
   Checksum = ReadByte(NO_ACK);
   Checksum = ReverseData(Checksum);
   
   /* Test checksum */
   CalcCRC = CRCTable[CalcCRC ^ CMD_STATUS_REG_R];
   CalcCRC = CRCTable[CalcCRC ^ Value];
   if (CalcCRC != Checksum)
   {
      Error = 1;
   } 
   
   *pValue = Value;
     
   return(Error);  
} /* ReadStatus */
#endif

/*************************************************************************/
/*  WriteStatus                                                          */
/*                                                                       */
/*  Writes the status register.                                          */
/*                                                                       */
/*  In    : Value                                                        */
/*  Out   : none                                                         */
/*  Return: 0 = no error / 1 = error                                     */
/*************************************************************************/
static uint8_t WriteStatus (uint8_t Value)
{
   uint8_t Error;
   
   TransStart();
   Error  = WriteByte(CMD_STATUS_REG_W);
   Error |= WriteByte(Value);
   
   return(Error);   
} /* WriteStatus */

/*=======================================================================*/
/*  All code exported                                                    */
/*=======================================================================*/

/*************************************************************************/
/*  sht15_Init                                                           */
/*                                                                       */
/*  In    : none                                                         */
/*  Out   : none                                                         */
/*  Return: none                                                         */
/*************************************************************************/
void sht15_Init (void)
{
   HWInit();
   
   COMReset();
   
   WriteStatus(0);
} /* sht15_Init */

/*************************************************************************/
/*  sht15_Measurement                                                    */
/*                                                                       */
/*  Return Temperature and Humidity.                                     */
/*                                                                       */
/*  In    : pTemperature, pHumidity                                      */
/*  Out   : pTemperature, pHumidity                                      */
/*  Return: 0 = no error / 1 = error                                     */
/*************************************************************************/
uint8_t sht15_Measurement (float *pTemperature, uint8_t *pHumidity)
{
   uint8_t  Error;
   uint16_t SO_T;
   uint16_t SO_RH;
   int16_t  Value;

   float    Temperature;
   float    Humidity;
   float    RH_lin;
   
   tal_LEDSet(TAL_LED_CHANNEL_2);

   Error  = ReadData(CMD_TEMP, &SO_T); 
   Error |= ReadData(CMD_HUMI, &SO_RH); 

   Temperature = -39.7f + (0.01f * SO_T);

   RH_lin   = -2.0468f + (0.0367f * SO_RH) + (-1.5955E-6f * SO_RH * SO_RH);
   Humidity = (Temperature - 25) * (0.01f + 0.00008f * SO_RH) + RH_lin;
      
   Value = (int16_t)(Temperature * 10);
   
   *pTemperature = (float)Value / 10;
   *pHumidity    = (uint8_t)(Humidity + 0.5);

   tal_LEDClear(TAL_LED_CHANNEL_2);

   return(Error);
} /* sht15_Measurement */

/*** EOF ***/
